/*
 * Copyright (c) 1999-2001 by Peter Simons <simons@ieee.org>.
 * All rights reserved.
 *
 * The routine parse_config_file() will parse a configuration file
 * from the file system and provide the results via a user-supplied
 * callback. It understands any sort of input file that obeys to the
 * following line format:
 *
 *     keyword data
 *
 * "keyword" and "data" have to be separated by one or more
 * white-space characters -- either tabs or blanks --, and "keyword"
 * has to begin at the first column of the line. The "data"-part may
 * be enclosed in quotation marks, but this is usually not necessary,
 * even if it contains whitespace itself. The only occasion, when the
 * data part must be quoted is in case it contains white-space as
 * first or last characters.
 *
 * Any line that starts with a '#' character as the first
 * non-whitespace-char is ignored and assumed to be a comment. Note
 * that comments after the data part are not recognized!
 *
 * Anything else is an "unknown" line and will be reported but not
 * parsed by the routine.
 *
 * parse_config_file() reports its findings to the caller via the
 * callback class "AbstractConfig", which is defined in this header.
 * The caller should derive his own version of this class and pass it
 * into parse_config_file() in order to process the data found by the
 * routine. In fact, AbstractConfig is pure virtual and cannot be
 * instantiated as it is.
 *
 * Look at the enclosed expample program "test.cc" for further
 * illustration.
 */

#ifndef PARSE_CONFIG_FILE_HPP
#define PARSE_CONFIG_FILE_HPP

// ISO C++ headers.
#include <string>

class AbstractConfig
    {
  public:
    virtual ~AbstractConfig() = 0;
  protected:
    virtual void set_option(const std::string& keyword, const std::string& data) = 0;
    virtual void unknown_line(const std::string& line) = 0;
    friend void parse_config_file(const char*, AbstractConfig&);
    };

void parse_config_file(const char* filename, AbstractConfig& config);

#endif // !defined(PARSE_CONFIG_FILE_HPP)
